package com.chopeks.glife

import antlr.KotlinLexer
import com.chopeks.qsp.QspType
import com.chopeks.transpiler.ExternalPreProcessor
import org.antlr.v4.runtime.*

/*
* Here's an example of simple API extension.
* In this case:
* 1. there's an enum defined, that describes original Girl Life locations in another file ->GLifeLocation.kt, or other enums, like pain or so
* 2. there are overloaded gs and gt methods for it
* 3. and lastly, a preprocessor that does translation from enum values into strings
*/

// 2. here are api overloads:
fun gs(location: Location, vararg parameter: QspType) = Unit

// 3. and preprocessor, note that this still has to output valid kotlin file
class GLifeEnumPreprocessor : ExternalPreProcessor {
  // let's help ourselves a bit
  private operator fun Token.rangeTo(other: Token) = startIndex..other.stopIndex

  private inline fun <reified T : Enum<T>> valueOf(name: Token) = "\"${enumValueOf<T>(name.text.trim('`')).name}\""

  override fun process(srcCode: String): String {
    // tokenize kotlin code
    val lexer = KotlinLexer(CharStreams.fromString(srcCode))
    val tokens = CommonTokenStream(lexer)
    // preprocess code
    var code = srcCode

    // Iterate from end of file, so it won't screw itself in case of many entries at once
    for (i in (tokens.numberOfOnChannelTokens - 1) downTo 2) {
      // so basically, an enum entry looks like this: `GLifeLocation.pavCommercial`
      // for example, Pavlovsk Commercial district is defined like: `pavCommercial("pavCommercial")`
      // what we wanna do, is translating enum entry `GLifeLocation.pavCommercial` into string `"pavCommercial"`
      // using lexer terms, `GLifeLocation.pavCommercial` is [id, dot, id]
      if (
        tokens[i].type == KotlinLexer.Identifier &&
        tokens[i - 1].type == KotlinLexer.DOT &&
        tokens[i - 2].type == KotlinLexer.Identifier
      ) {
        // in case of possible enum entry found, let's replace
        // from start of 1st Identifier up to end of 2nd Identifier put real name of file

        val name = if (tokens[i].text.startsWith("`")) tokens[i].text.trim('`') else tokens[i].text
        code = when (tokens[i - 2].text) {
          GLLoc::class.simpleName -> {
            var j = i - 2
            while (tokens[j].type == KotlinLexer.DOT || tokens[j].type == KotlinLexer.Identifier) j++
            val callDef = code.substring(tokens[i - 2]..tokens[j - 1])
            val clazz = Class.forName("com.chopeks.glife." + callDef.replace(".", "$").replace("`", "")).kotlin.objectInstance as Location
            when (tokens[i - 4].text) {
              "gt", "gs", "xgt" -> {
                val split = callDef.split(".").toMutableList().also {
                  it.removeAt(it.size - 1)
                }
                if (split.size == 1) {
                  // that means only GLLoc left
                  code.replaceRange(tokens[i - 2]..tokens[j - 1], "\"${clazz.name}\"")
                } else {
                  val callDef = code.substring(tokens[i - 2]..tokens[i])
                  val clazz2 = Class.forName("com.chopeks.glife." + callDef.replace(".", "$").replace("`", "")).kotlin.objectInstance as Location
                  code.replaceRange(tokens[i - 2]..tokens[j - 1], "\"${clazz2.name}\", \"${clazz.name}\"")
                }
              }
              else -> code.replaceRange(tokens[i - 2]..tokens[j - 1], "\"${clazz.name}\"")
            }
          }
          GLNpc::class.simpleName -> code.replaceRange(tokens[i - 2]..tokens[i], "\"${GLNpc.valueOf(name).lName}\"")
          else -> {
            try {
              val clazz = Class.forName("com.chopeks.glife." + tokens[i - 2].text).enumConstants as Array<Enum<*>>
              val value = clazz.first { it.name == tokens[i].text }
              code.replaceRange(tokens[i - 2]..tokens[i], "\"${value.name}\"")
            } catch (e: Throwable) {
              code
            }
          }
        }
      }
    }
    return code
  }
}
// NOTE: Do not forget to add your preprocessor to transpiler!
